@@ -3,10 +3,12 @@ import os
 import re
 from datetime import datetime, timedelta
 from typing import (
+    Any,
     AsyncContextManager,
     AsyncIterator,
     Generic,
     Iterable,
+    Iterator,
     List,
     Mapping,
     Optional,
@@ -15,6 +17,7 @@ from typing import (
     Union,
     cast,
 )
+from urllib.parse import quote, unquote
 
 import click
 from click import BadParameter
@@ -114,35 +117,122 @@ class LocalImageType(click.ParamType):
         return client.parse.local_image(value)
 
 
-class RemoteTaglessImageType(click.ParamType):
-    name = "image"
-
-    def convert(
-        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
-    ) -> RemoteImage:
-        assert ctx is not None
-        root = cast(Root, ctx.obj)
-        client = root.run(root.init_client())
-        return client.parse.remote_image(value, tag_option=TagOption.DENY)
+def _complete_clusters(
+    client: Client,
+    prefix: str,
+    incomplete: str,
+) -> List[CompletionItem]:
+    return [
+        CompletionItem(f"{name}/", type="uri", prefix=prefix)
+        for name in client.config.clusters
+        if name.startswith(incomplete)
+    ]
 
 
-class RemoteImageType(click.ParamType):
+class RemoteImageType(AsyncType[RemoteImage]):
     name = "image"
 
     def __init__(self, tag_option: TagOption = TagOption.DEFAULT) -> None:
         self.tag_option = tag_option
 
-    def convert(
-        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
+    async def async_convert(
+        self,
+        root: Root,
+        value: str,
+        param: Optional[click.Parameter],
+        ctx: Optional[click.Context],
     ) -> RemoteImage:
-        assert ctx is not None
-        root = cast(Root, ctx.obj)
-        cluster_name = ctx.params.get("cluster")
-        client = root.run(root.init_client())
+        client = await root.init_client()
+        cluster_name = client.cluster_name
+        if ctx:
+            cluster_name = ctx.params.get("cluster", client.cluster_name)
         return client.parse.remote_image(
             value, tag_option=self.tag_option, cluster_name=cluster_name
         )
 
+    async def _complete_image_names(
+        self,
+        client: Client,
+        uri_prefix: str,
+        path_prefix: str,
+        cluster_name: str,
+        incomplete: str,
+    ) -> List[CompletionItem]:
+        if cluster_name not in client.config.clusters:
+            return []
+        names = []
+        prefix = f"{path_prefix}{unquote(incomplete)}"
+        for image in await client.images.list(cluster_name):
+            path = f"{image.owner}/{image.name}"
+            if image.org_name:
+                path = f"{image.org_name}/{path}"
+            if path.startswith(prefix):
+                names.append(incomplete + quote(path[len(prefix) :]))
+        return [CompletionItem(name, type="uri", prefix=uri_prefix) for name in names]
+
+    async def _complete_image_tags(
+        self,
+        client: Client,
+        image_str: str,
+        incomplete: str,
+    ) -> List[CompletionItem]:
+        image = client.parse.remote_image(image_str, tag_option=TagOption.DENY)
+        result = []
+        for image_tag in await client.images.tags(image):
+            assert image_tag.tag
+            if image_tag.tag.startswith(incomplete):
+                result.append(
+                    CompletionItem(image_tag.tag, type="uri", prefix=image_str + ":")
+                )
+        return result
+
+    async def async_shell_complete(
+        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
+    ) -> List[CompletionItem]:
+        if "image".startswith(incomplete):
+            return [CompletionItem("image:", type="uri", prefix="")]
+
+        async with await root.init_client() as client:
+            if incomplete.startswith("image:"):
+                incomplete = incomplete[len("image:") :]
+                if self.tag_option != TagOption.DENY and ":" in incomplete:
+                    prefix, incomplete = incomplete.split(":", 1)
+                    return await self._complete_image_tags(
+                        client, f"image:{prefix}", incomplete
+                    )
+
+                if incomplete.startswith("///"):
+                    return []
+
+                if incomplete.startswith("//"):
+                    incomplete = incomplete[2:]
+                    if "/" not in incomplete:
+                        return _complete_clusters(client, "image://", incomplete)
+                    cluster_name, incomplete = incomplete.split("/", 1)
+                    return await self._complete_image_names(
+                        client, f"image://{cluster_name}/", "", cluster_name, incomplete
+                    )
+
+                if incomplete.startswith("/"):
+                    incomplete = incomplete[1:]
+                    return await self._complete_image_names(
+                        client, "image:/", "", client.cluster_name, incomplete
+                    )
+
+                path_prefix = f"{client.username}/"
+                if client.config.org_name:
+                    path_prefix = f"{client.config.org_name}/{path_prefix}"
+                return await self._complete_image_names(
+                    client, "image:", path_prefix, client.cluster_name, incomplete
+                )
+
+            return []
+
+
+class RemoteTaglessImageType(RemoteImageType):
+    def __init__(self, tag_option: TagOption = TagOption.DENY) -> None:
+        super().__init__(tag_option=TagOption.DENY)
+
 
 class LocalRemotePortParamType(click.ParamType):
     name = "local-remote-port-pair"
@@ -170,7 +260,7 @@ class MegabyteType(click.ParamType):
         self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
     ) -> int:
         if isinstance(value, int):
-            return int(value / (1024 ** 2))
+            return int(value / (1024**2))
         return to_megabytes(value)
 
 
@@ -438,34 +528,123 @@ class JobType(AsyncType[str]):
     ) -> str:
         return value
 
+    async def _complete_job_owners(
+        self,
+        client: Client,
+        prefix: str,
+        cluster_name: str,
+        incomplete: str,
+    ) -> List[CompletionItem]:
+        if cluster_name not in client.config.clusters:
+            return []
+        now = datetime.now()
+        limit = int(os.environ.get(JOB_LIMIT_ENV, 100))
+        names = []
+        async with client.jobs.list(
+            since=now - timedelta(days=7),
+            reverse=True,
+            limit=limit,
+            cluster_name=cluster_name,
+        ) as it:
+            async for job in it:
+                name = job.owner
+                if name.startswith(incomplete) and name not in names:
+                    names.append(name)
+        return [CompletionItem(f"{name}/", type="uri", prefix=prefix) for name in names]
+
+    async def _complete_job_names(
+        self,
+        client: Client,
+        prefix: str,
+        cluster_name: str,
+        username: str,
+        incomplete: str,
+    ) -> List[CompletionItem]:
+        if cluster_name not in client.config.clusters:
+            return []
+        now = datetime.now()
+        limit = int(os.environ.get(JOB_LIMIT_ENV, 100))
+        names = {}
+        async with client.jobs.list(
+            since=now - timedelta(days=7),
+            reverse=True,
+            limit=limit,
+            cluster_name=cluster_name,
+            owners=[username],
+        ) as it:
+            async for job in it:
+                id = job.id
+                name = job.name
+                if id.startswith(incomplete):
+                    names[id] = name
+                if name and name.startswith(incomplete):
+                    names[name] = id
+        if prefix:
+            return [CompletionItem(name, type="uri", prefix=prefix) for name in names]
+        else:
+            return [CompletionItem(name, help=help) for name, help in names.items()]
+
     async def async_shell_complete(
         self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
     ) -> List[CompletionItem]:
-        async with await root.init_client() as client:
-            ret: List[CompletionItem] = []
-            now = datetime.now()
-            limit = int(os.environ.get(JOB_LIMIT_ENV, 100))
-            async with client.jobs.list(
-                since=now - timedelta(days=7), reverse=True, limit=limit
-            ) as it:
-                async for job in it:
-                    job_name = job.name or ""
-                    for test in (
-                        job.id,
-                        job_name,
-                        f"job:{job.id}",
-                        f"job:/{job.owner}/{job.id}",
-                        f"job://{job.cluster_name}/{job.owner}/{job.id}",
-                    ):
-                        if test.startswith(incomplete):
-                            ret.append(CompletionItem(test, help=job_name))
+        if "job".startswith(incomplete):
+            return [CompletionItem("job:", type="uri", prefix="")]
 
-            return ret
+        async with await root.init_client() as client:
+            if incomplete.startswith("job:///"):
+                return []
+
+            if incomplete.startswith("job://"):
+                parts = incomplete[len("job://") :].split("/")
+                if len(parts) == 1:
+                    return _complete_clusters(client, "job://", *parts)
+                elif len(parts) == 2:
+                    return await self._complete_job_owners(
+                        client, f"job://{parts[0]}/", *parts
+                    )
+                elif len(parts) == 3:
+                    return await self._complete_job_names(
+                        client, f"job://{parts[0]}/{parts[1]}/", *parts
+                    )
+                return []
+
+            if incomplete.startswith("job:/"):
+                parts = incomplete[len("job:/") :].split("/")
+                if len(parts) == 1:
+                    return await self._complete_job_owners(
+                        client, "job:/", client.cluster_name, *parts
+                    )
+                elif len(parts) == 2:
+                    return await self._complete_job_names(
+                        client, f"job:/{parts[0]}/", client.cluster_name, *parts
+                    )
+                return []
+
+            if incomplete.startswith("job:"):
+                parts = incomplete[len("job:") :].split("/")
+                if len(parts) == 1:
+                    return await self._complete_job_names(
+                        client, "job:", client.cluster_name, client.username, *parts
+                    )
+                return []
+
+            return await self._complete_job_names(
+                client, "", client.cluster_name, client.username, incomplete
+            )
 
 
 JOB = JobType()
 
 
+def _complete_id_name(
+    id: str, name: Optional[str], incomplete: str
+) -> Iterator[CompletionItem]:
+    if id.startswith(incomplete):
+        yield CompletionItem(id, help=name or "")
+    if name and name.startswith(incomplete):
+        yield CompletionItem(name, help=id)
+
+
 class DiskType(AsyncType[str]):
     name = "disk"
 
@@ -485,14 +664,7 @@ class DiskType(AsyncType[str]):
             ret: List[CompletionItem] = []
             async with client.disks.list(cluster_name=ctx.params.get("cluster")) as it:
                 async for disk in it:
-                    disk_name = disk.name or ""
-                    for test in (
-                        disk.id,
-                        disk_name,
-                    ):
-                        if test.startswith(incomplete):
-                            ret.append(CompletionItem(test, help=disk_name))
-
+                    ret.extend(_complete_id_name(disk.id, disk.name, incomplete))
             return ret
 
 
@@ -500,7 +672,7 @@ DISK = DiskType()
 
 
 class ServiceAccountType(AsyncType[str]):
-    name = "disk"
+    name = "service_account"
 
     async def async_convert(
         self,
@@ -518,13 +690,7 @@ class ServiceAccountType(AsyncType[str]):
             ret: List[CompletionItem] = []
             async with client.service_accounts.list() as it:
                 async for account in it:
-                    account_name = account.name or ""
-                    for test in (
-                        account.id,
-                        account_name,
-                    ):
-                        if test.startswith(incomplete):
-                            ret.append(CompletionItem(test, help=account_name))
+                    ret.extend(_complete_id_name(account.id, account.name, incomplete))
 
             return ret
 
@@ -560,10 +726,10 @@ class PathURLCompleter(URLCompleter, abc.ABC):
         self._complete_dir = complete_dir
         self._complete_file = complete_file
 
-    def _get_dir_uri(self, uri: URL, incomplete: str) -> URL:
-        if incomplete.endswith("/"):
-            return uri / ""
-        return uri.parent / ""
+    def _split_uri(self, uri: URL, incomplete: str) -> Tuple[URL, str]:
+        if incomplete.endswith("/") or incomplete == uri.scheme + ":":
+            return uri, ""
+        return uri.parent, uri.name
 
     def _make_item(
         self,
@@ -576,7 +742,7 @@ class PathURLCompleter(URLCompleter, abc.ABC):
         return CompletionItem(
             name,
             type="uri",
-            prefix=str(parent) + "/" if parent.path else str(parent),
+            prefix=str(parent / ""),
         )
 
     @abc.abstractmethod
@@ -595,18 +761,19 @@ class PathURLCompleter(URLCompleter, abc.ABC):
         root: Root,
         incomplete: str,
     ) -> AsyncIterator[CompletionItem]:
-        dir_uri = self._get_dir_uri(uri, incomplete)
+        dir_uri, incomplete_name = self._split_uri(uri, incomplete)
         if not await self._is_valid_dir(root, dir_uri):
             return
         async with self._iter_dir(root, dir_uri) as it:
             async for item in it:
-                if str(dir_uri / item.name).startswith(incomplete):
+                if item.name.startswith(incomplete_name):
                     if item.is_dir() and not self._complete_dir:
                         continue
                     if not item.is_dir() and not self._complete_file:
                         continue
+
                     yield self._make_item(
-                        dir_uri.parent,
+                        dir_uri,
                         item.name,
                         item.is_dir(),
                     )
@@ -645,20 +812,16 @@ class StoragePathURLCompleter(PathURLCompleter):
 
 class BlobPathURLCompleter(PathURLCompleter):
     async def _is_valid_dir(self, root: Root, uri: URL) -> bool:
-        try:
-            return await root.client.buckets.blob_is_dir(uri)
-        except ResourceNotFound:
-            return False
+        # Not used
+        raise NotImplementedError
 
     @asyncgeneratorcontextmanager
     async def _iter_dir(
         self, root: Root, uri: URL
     ) -> AsyncIterator[PathURLCompleter.DirEntry]:
-        async with root.client.buckets.list_blobs(uri) as it:
-            async for blob_entry in it:
-                if blob_entry.uri == uri:
-                    continue  # Directory itself is also listed as it is prefix search
-                yield blob_entry
+        # Not used
+        raise NotImplementedError
+        yield
 
     async def get_completions(
         self,
@@ -680,14 +843,31 @@ class BlobPathURLCompleter(PathURLCompleter):
                     names = [bucket.id] + ([bucket.name] if bucket.name else [])
                     for name in names:
                         if str(prefix / name).startswith(incomplete):
-                            yield self._make_item(
-                                prefix,
-                                name,
-                                True,
-                            )
+                            yield self._make_item(prefix, name, True)
         else:
-            async for item in super().get_completions(uri, root, incomplete):
-                yield item
+            # Generic get_completions() is not used here because we can
+            # benefit from prefix search in list_blobs().
+            if incomplete.endswith("/"):
+                prefix = uri
+                full_uri = full_uri / ""
+                skip_uri_len = len(full_uri.parts)
+            else:
+                prefix = uri.parent
+                skip_uri_len = None
+
+            async with root.client.buckets.list_blobs(full_uri) as it:
+                async for item in it:
+                    if item.is_dir():
+                        if not self._complete_dir:
+                            continue
+                        # Directory itself is also listed as it is prefix search
+                        if len(item.uri.parts) == skip_uri_len:
+                            continue
+                    else:
+                        if not self._complete_file:
+                            continue
+
+                    yield self._make_item(prefix, item.name, item.is_dir())
 
 
 class PlatformURIType(AsyncType[URL]):
@@ -735,24 +915,25 @@ class PlatformURIType(AsyncType[URL]):
         else:
             return ret
 
+        if scheme != "file:" and incomplete == scheme + "//":
+            return _complete_clusters(root.client, incomplete, "")
+
         uri = root.client.parse.str_to_uri(
             incomplete,
             allowed_schemes=self._allowed_schemes,
-            short=not ("://" in incomplete),
+            short=not (
+                incomplete.startswith(scheme + "//")
+                and not incomplete.startswith(scheme + "///")
+            ),
         )
-        if uri.host and uri.path == "/" and not incomplete.endswith("/"):
+        if (
+            uri.scheme != "file"
+            and uri.host
+            and uri.path == "/"
+            and not incomplete.endswith("/")
+        ):
             # Cluster name is incomplete
-            for cluster_name in root.client.config.clusters:
-                if cluster_name.startswith(uri.host):
-                    ret.append(
-                        CompletionItem(
-                            f"{cluster_name}/", type="uri", prefix=f"{uri.scheme}://"
-                        )
-                    )
-            return ret
-        if uri.host and uri.path.count("/") == 1:
-            # Username completion not supported
-            return []
+            return _complete_clusters(root.client, f"{uri.scheme}://", uri.host)
         completer = self._completers.get(uri.scheme)
         if completer:
             return [
@@ -923,13 +1104,7 @@ class BucketType(AsyncType[str]):
                 cluster_name=ctx.params.get("cluster")
             ) as it:
                 async for bucket in it:
-                    bucket_name = bucket.name or ""
-                    for test in (
-                        bucket.id,
-                        bucket_name,
-                    ):
-                        if test.startswith(incomplete):
-                            ret.append(CompletionItem(test, help=bucket_name))
+                    ret.extend(_complete_id_name(bucket.id, bucket.name, incomplete))
 
             return ret
 
@@ -958,13 +1133,9 @@ class BucketCredentialType(AsyncType[str]):
                 cluster_name=ctx.params.get("cluster")
             ) as it:
                 async for credential in it:
-                    credential_name = credential.name or ""
-                    for test in (
-                        credential.id,
-                        credential_name,
-                    ):
-                        if test.startswith(incomplete):
-                            ret.append(CompletionItem(test, help=credential_name))
+                    ret.extend(
+                        _complete_id_name(credential.id, credential.name, incomplete)
+                    )
 
             return ret
 
@@ -972,6 +1143,36 @@ class BucketCredentialType(AsyncType[str]):
 BUCKET_CREDENTIAL = BucketCredentialType()
 
 
+class UnionType(AsyncType[Any]):
+    def __init__(self, name: str, *types: AsyncType[Any]) -> None:
+        self.name = name
+        self._inner_types = types
+
+    async def async_convert(
+        self,
+        root: Root,
+        value: str,
+        param: Optional[click.Parameter],
+        ctx: Optional[click.Context],
+    ) -> Any:
+        for inner_type in self._inner_types:
+            try:
+                return await inner_type.async_convert(root, value, param, ctx)
+            except ValueError:
+                pass
+        raise ValueError(f"Cannot parse {value} as {self.name}")
+
+    async def async_shell_complete(
+        self, root: Root, ctx: click.Context, param: click.Parameter, incomplete: str
+    ) -> List[CompletionItem]:
+        result = []
+        for inner_type in self._inner_types:
+            result += await inner_type.async_shell_complete(
+                root, ctx, param, incomplete
+            )
+        return result
+
+
 def setup_shell_completion() -> None:
     add_completion_class(NewZshComplete)
     add_completion_class(NewBashComplete)
