@@ -12,13 +12,18 @@ from neuro_sdk import (
     _ClusterUserWithInfo,
     _ConfigCluster,
     _NodePool,
+    _OrgCluster,
     _Quota,
     _Storage,
     _StorageInstance,
     _UserInfo,
 )
 
-from neuro_cli.formatters.admin import ClustersFormatter, ClusterUserFormatter
+from neuro_cli.formatters.admin import (
+    ClustersFormatter,
+    ClusterUserFormatter,
+    OrgClusterFormatter,
+)
 
 RichCmp = Callable[[RenderableType], None]
 
@@ -132,6 +137,7 @@ class TestClustersFormatter:
                     name="default",
                     default_credits=Decimal(20),
                     default_quota=_Quota(total_running_jobs=42),
+                    default_role=_ClusterUserRoleType.USER,
                 ),
                 _ConfigCluster(name="default", status="deployed"),
             )
@@ -142,7 +148,12 @@ class TestClustersFormatter:
         formatter = ClustersFormatter()
         clusters = {
             "on-prem": (
-                _Cluster(name="on-prem", default_credits=None, default_quota=_Quota()),
+                _Cluster(
+                    name="on-prem",
+                    default_credits=None,
+                    default_quota=_Quota(),
+                    default_role=_ClusterUserRoleType.USER,
+                ),
                 _ConfigCluster(
                     name="on-prem",
                     status="deployed",
@@ -162,7 +173,12 @@ class TestClustersFormatter:
         formatter = ClustersFormatter()
         clusters = {
             "default": (
-                _Cluster(name="default", default_credits=None, default_quota=_Quota()),
+                _Cluster(
+                    name="default",
+                    default_credits=None,
+                    default_quota=_Quota(),
+                    default_role=_ClusterUserRoleType.USER,
+                ),
                 _ConfigCluster(
                     name="default",
                     status="deployed",
@@ -190,7 +206,12 @@ class TestClustersFormatter:
         formatter = ClustersFormatter()
         clusters = {
             "default": (
-                _Cluster(name="default", default_credits=None, default_quota=_Quota()),
+                _Cluster(
+                    name="default",
+                    default_credits=None,
+                    default_quota=_Quota(),
+                    default_role=_ClusterUserRoleType.USER,
+                ),
                 _ConfigCluster(
                     name="default",
                     status="deployed",
@@ -218,7 +239,12 @@ class TestClustersFormatter:
         formatter = ClustersFormatter()
         clusters = {
             "default": (
-                _Cluster(name="default", default_credits=None, default_quota=_Quota()),
+                _Cluster(
+                    name="default",
+                    default_credits=None,
+                    default_quota=_Quota(),
+                    default_role=_ClusterUserRoleType.USER,
+                ),
                 _ConfigCluster(
                     name="default",
                     status="deployed",
@@ -245,7 +271,12 @@ class TestClustersFormatter:
         formatter = ClustersFormatter()
         clusters = {
             "default": (
-                _Cluster(name="default", default_credits=None, default_quota=_Quota()),
+                _Cluster(
+                    name="default",
+                    default_credits=None,
+                    default_quota=_Quota(),
+                    default_role=_ClusterUserRoleType.USER,
+                ),
                 _ConfigCluster(
                     name="default",
                     status="deployed",
@@ -263,3 +294,25 @@ class TestClustersFormatter:
             )
         }
         rich_cmp(formatter(clusters))
+
+
+class TestOrgClusterFormatter:
+    def test_org_cluster_formatter(self, rich_cmp: RichCmp) -> None:
+        formatter = OrgClusterFormatter()
+        cluster = _OrgCluster(
+            cluster_name="test",
+            org_name="test-org",
+            quota=_Quota(total_running_jobs=2),
+            balance=_Balance(credits=Decimal(100), spent_credits=Decimal(20)),
+        )
+        rich_cmp(formatter(cluster))
+
+    def test_org_cluster_formatter_no_quota(self, rich_cmp: RichCmp) -> None:
+        formatter = OrgClusterFormatter()
+        cluster = _OrgCluster(
+            cluster_name="test",
+            org_name="test-org",
+            quota=_Quota(),
+            balance=_Balance(),
+        )
+        rich_cmp(formatter(cluster))
