@@ -1,6 +1,6 @@
 from datetime import timedelta
 from pathlib import Path
-from typing import Any, Callable, Dict, NoReturn
+from typing import Any, Callable, Dict, List, NoReturn, Optional, Set, Union, cast
 from unittest import mock
 
 import click
@@ -28,12 +28,22 @@ from tests import _TestServerFactory
 _MakeClient = Callable[..., Client]
 
 
-def _job_entry(job_id: str, cluster_name: str = "default") -> Dict[str, Any]:
+def _job_entry(
+    job_id: str,
+    cluster_name: str = "default",
+    owner: str = "user",
+    org_name: Optional[str] = None,
+) -> Dict[str, Any]:
+    if org_name:
+        uri = f"job://{cluster_name}/{org_name}/{owner}/{job_id}"
+    else:
+        uri = f"job://{cluster_name}/{owner}/{job_id}"
     return {
         "id": job_id,
-        "owner": "job-owner",
+        "owner": owner,
         "cluster_name": cluster_name,
-        "uri": f"job://{cluster_name}/job-owner/{job_id}",
+        "org_name": org_name,
+        "uri": uri,
         "status": "running",
         "history": {
             "status": "running",
@@ -53,18 +63,25 @@ def _job_entry(job_id: str, cluster_name: str = "default") -> Dict[str, Any]:
         "scheduler_enabled": True,
         "pass_config": False,
         "name": "my-job-name",
-        "internal_hostname": f"job-id.{cluster_name}",
-        "internal_hostname_named": f"my-job-name--job-owner.{cluster_name}",
+        "internal_hostname": f"{job_id}.{cluster_name}",
+        "internal_hostname_named": f"my-job-name--{owner}.{cluster_name}",
         "total_price_credits": "150",
         "price_credits_per_hour": "15",
     }
 
 
-def _check_params(request: web.Request, **expected: str) -> None:
+def _check_params(
+    request: web.Request, **expected: Union[str, List[str], Set[str]]
+) -> None:
     # Since `resolve_job` excepts any Exception, `assert` will be caught there
     for key, value in expected.items():
-        got = request.query.get(key)
+        got: Union[List[str], Set[str]] = request.query.getall(key)
+        if isinstance(value, set):
+            got = set(got)
+        elif not isinstance(value, list):
+            value = [value]
         if got != value:
+            print(f"received: {key}={got!r} (expected {value!r})")
             pytest.fail(f"received: {key}={got!r} (expected {value!r})")
 
 
@@ -81,7 +98,6 @@ async def test_resolve_job_id__from_string__no_jobs_found(
             owner="user",
             cluster_name="default",
             reverse="1",
-            limit="1",
         )
         return web.json_response(JSON)
 
@@ -115,7 +131,6 @@ async def test_resolve_job_id__from_uri_with_owner_same_user__no_jobs_found(
             owner=job_owner,
             cluster_name=cluster_name,
             reverse="1",
-            limit="1",
         )
         return web.json_response(JSON)
 
@@ -148,7 +163,6 @@ async def test_resolve_job_id__from_uri_with_owner_other_user__no_jobs_found(
             owner=job_owner,
             cluster_name="default",
             reverse="1",
-            limit="1",
         )
         return web.json_response(JSON)
 
@@ -178,7 +192,6 @@ async def test_resolve_job_id__from_uri_without_owner__no_jobs_found(
             owner="user",
             cluster_name="default",
             reverse="1",
-            limit="1",
         )
         return web.json_response(JSON)
 
@@ -201,7 +214,7 @@ async def test_resolve_job_id__from_string__single_job_found(
 ) -> None:
     job_name = "test-job-name-555"
     job_id = "job-id-1"
-    JSON = {"jobs": [_job_entry(job_id)]}
+    JSON = {"jobs": [_job_entry(job_id, owner="user")]}
 
     async def handler(request: web.Request) -> web.Response:
         _check_params(
@@ -210,7 +223,6 @@ async def test_resolve_job_id__from_string__single_job_found(
             owner="user",
             cluster_name="default",
             reverse="1",
-            limit="1",
         )
         return web.json_response(JSON)
 
@@ -231,23 +243,32 @@ async def test_resolve_job_id__from_string__single_job_found(
 
 
 @pytest.mark.parametrize("cluster_name", ["default", "other"])
+@pytest.mark.parametrize("org_name", [None, "test-org"])
+@pytest.mark.parametrize(
+    "job_owner", ["job-owner", "user", "job-owner/service", "user/service"]
+)
 async def test_resolve_job_id__from_uri_with_owner__single_job_found(
-    aiohttp_server: _TestServerFactory, make_client: _MakeClient, cluster_name: str
+    aiohttp_server: _TestServerFactory,
+    make_client: _MakeClient,
+    cluster_name: str,
+    org_name: Optional[str],
+    job_owner: str,
 ) -> None:
-    job_owner = "job-owner"
     job_name = "my-job-name"
     uri = f"job://{cluster_name}/{job_owner}/{job_name}"
     job_id = "job-id-1"
-    JSON = {"jobs": [_job_entry(job_id, cluster_name=cluster_name)]}
+    JSON = {"jobs": [_job_entry(job_id, cluster_name=cluster_name, owner=job_owner)]}
 
     async def handler(request: web.Request) -> web.Response:
+        expected_owners = (
+            [job_owner, "service"] if job_owner == "job-owner/service" else [job_owner]
+        )
         _check_params(
             request,
             name=job_name,
-            owner=job_owner,
+            owner=expected_owners,
             cluster_name=cluster_name,
             reverse="1",
-            limit="1",
         )
         return web.json_response(JSON)
 
@@ -256,7 +277,7 @@ async def test_resolve_job_id__from_uri_with_owner__single_job_found(
 
     srv = await aiohttp_server(app)
 
-    async with make_client(srv.make_url("/")) as client:
+    async with make_client(srv.make_url("/"), org_name=org_name) as client:
         resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
         assert resolved == job_id
         resolved_ex = await resolve_job_ex(
@@ -265,22 +286,42 @@ async def test_resolve_job_id__from_uri_with_owner__single_job_found(
         assert resolved_ex == (job_id, cluster_name)
 
 
-async def test_resolve_job_id__from_uri_without_owner__single_job_found(
-    aiohttp_server: _TestServerFactory, make_client: _MakeClient
+@pytest.mark.parametrize("cluster_name", ["default", "other"])
+@pytest.mark.parametrize("org_name", [None, "test-org", "job-org"])
+@pytest.mark.parametrize(
+    "job_owner", ["job-owner", "user", "job-owner/service", "user/service"]
+)
+async def test_resolve_job_id__from_uri_with_org__single_job_found(
+    aiohttp_server: _TestServerFactory,
+    make_client: _MakeClient,
+    cluster_name: str,
+    org_name: Optional[str],
+    job_owner: str,
 ) -> None:
+    job_org = "job-org"
     job_name = "my-job-name"
-    uri = f"job:{job_name}"
+    uri = f"job://{cluster_name}/{job_org}/{job_owner}/{job_name}"
     job_id = "job-id-1"
-    JSON = {"jobs": [_job_entry(job_id)]}
+    JSON = {
+        "jobs": [
+            _job_entry(
+                job_id, cluster_name=cluster_name, owner=job_owner, org_name=job_org
+            )
+        ]
+    }
 
     async def handler(request: web.Request) -> web.Response:
+        expected_owners = (
+            [f"{job_org}/{job_owner}", job_owner]
+            if org_name != job_org
+            else [job_owner]
+        )
         _check_params(
             request,
             name=job_name,
-            owner="user",
-            cluster_name="default",
+            owner=expected_owners,
+            cluster_name=cluster_name,
             reverse="1",
-            limit="1",
         )
         return web.json_response(JSON)
 
@@ -289,7 +330,158 @@ async def test_resolve_job_id__from_uri_without_owner__single_job_found(
 
     srv = await aiohttp_server(app)
 
-    async with make_client(srv.make_url("/")) as client:
+    async with make_client(srv.make_url("/"), org_name=org_name) as client:
+        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
+        assert resolved == job_id
+        resolved_ex = await resolve_job_ex(
+            uri, client=client, status={JobStatus.RUNNING}
+        )
+        assert resolved_ex == (job_id, cluster_name)
+
+
+@pytest.mark.parametrize("cluster_name", ["default", "other"])
+@pytest.mark.parametrize("org_name", [None, "test-org"])
+async def test_resolve_job_id__from_uri__multiple_jobs_found(
+    aiohttp_server: _TestServerFactory,
+    make_client: _MakeClient,
+    cluster_name: str,
+    org_name: Optional[str],
+) -> None:
+    job_org = f"job-owner"
+    job_owner = f"job-owner"
+    job_name = "my-job-name"
+    uri = f"job://{cluster_name}/{job_owner}/{job_name}"
+    job_ids = [f"job-id-{i}" for i in range(4)]
+    JSON = {
+        "jobs": [
+            _job_entry(
+                job_ids[0],
+                cluster_name=cluster_name,
+                owner=job_owner,
+                org_name="other-org",
+            ),
+            _job_entry(
+                job_ids[1], cluster_name=cluster_name, owner=job_owner, org_name=job_org
+            ),
+            _job_entry(job_ids[2], cluster_name=cluster_name, owner=job_owner),
+            _job_entry(
+                job_ids[3],
+                cluster_name=cluster_name,
+                owner=job_owner,
+                org_name="other-org",
+            ),
+        ]
+    }
+
+    async def handler(request: web.Request) -> web.Response:
+        _check_params(
+            request,
+            name=job_name,
+            owner=job_owner,
+            cluster_name=cluster_name,
+            reverse="1",
+        )
+        return web.json_response(JSON)
+
+    app = web.Application()
+    app.router.add_get("/jobs", handler)
+
+    srv = await aiohttp_server(app)
+
+    async with make_client(srv.make_url("/"), org_name=org_name) as client:
+        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
+        assert resolved == job_ids[2]
+        resolved_ex = await resolve_job_ex(
+            uri, client=client, status={JobStatus.RUNNING}
+        )
+        assert resolved_ex == (job_ids[2], cluster_name)
+
+
+@pytest.mark.parametrize("cluster_name", ["default", "other"])
+@pytest.mark.parametrize("org_name", [None, "test-org"])
+async def test_resolve_job_id__from_uri_with_org__multiple_jobs_found(
+    aiohttp_server: _TestServerFactory,
+    make_client: _MakeClient,
+    cluster_name: str,
+    org_name: Optional[str],
+) -> None:
+    job_org = f"job-owner"
+    job_owner = f"job-owner"
+    job_name = "my-job-name"
+    uri = f"job://{cluster_name}/{job_org}/{job_owner}/{job_name}"
+    job_ids = [f"job-id-{i}" for i in range(4)]
+    JSON = {
+        "jobs": [
+            _job_entry(
+                job_ids[0],
+                cluster_name=cluster_name,
+                owner=job_owner,
+                org_name="other-org",
+            ),
+            _job_entry(job_ids[1], cluster_name=cluster_name, owner=job_owner),
+            _job_entry(
+                job_ids[2], cluster_name=cluster_name, owner=job_owner, org_name=job_org
+            ),
+            _job_entry(
+                job_ids[3],
+                cluster_name=cluster_name,
+                owner=job_owner,
+                org_name="other-org",
+            ),
+        ]
+    }
+
+    async def handler(request: web.Request) -> web.Response:
+        _check_params(
+            request,
+            name=job_name,
+            owner=[f"{job_org}/{job_owner}", job_owner],
+            cluster_name=cluster_name,
+            reverse="1",
+        )
+        return web.json_response(JSON)
+
+    app = web.Application()
+    app.router.add_get("/jobs", handler)
+
+    srv = await aiohttp_server(app)
+
+    async with make_client(srv.make_url("/"), org_name=org_name) as client:
+        resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
+        assert resolved == job_ids[2]
+        resolved_ex = await resolve_job_ex(
+            uri, client=client, status={JobStatus.RUNNING}
+        )
+        assert resolved_ex == (job_ids[2], cluster_name)
+
+
+@pytest.mark.parametrize("org_name", [None, "test-org"])
+async def test_resolve_job_id__from_uri_without_owner__single_job_found(
+    aiohttp_server: _TestServerFactory,
+    make_client: _MakeClient,
+    org_name: Optional[str],
+) -> None:
+    job_name = "my-job-name"
+    uri = f"job:{job_name}"
+    job_id = "job-id-1"
+    JSON = {"jobs": [_job_entry(job_id, owner="user", org_name=org_name)]}
+
+    async def handler(request: web.Request) -> web.Response:
+        _check_params(
+            request,
+            name=job_name,
+            owner="user",
+            cluster_name="default",
+            reverse="1",
+        )
+        return web.json_response(JSON)
+
+    app = web.Application()
+    app.router.add_get("/jobs", handler)
+
+    srv = await aiohttp_server(app)
+
+    async with make_client(srv.make_url("/"), org_name=org_name) as client:
         resolved = await resolve_job(uri, client=client, status={JobStatus.RUNNING})
         assert resolved == job_id
         resolved_ex = await resolve_job_ex(
@@ -311,7 +503,6 @@ async def test_resolve_job_id__server_error(
             owner="user",
             cluster_name="default",
             reverse="1",
-            limit="1",
         )
         raise web.HTTPError()
 
@@ -346,7 +537,6 @@ async def test_resolve_job_id__from_uri_with_owner_same_user__server_error(
             owner=job_owner,
             cluster_name=cluster_name,
             reverse="1",
-            limit="1",
         )
         raise web.HTTPError()
 
@@ -378,7 +568,6 @@ async def test_resolve_job_id__from_uri_with_owner_other_user__server_error(
             owner=job_owner,
             cluster_name="default",
             reverse="1",
-            limit="1",
         )
         raise web.HTTPError()
 
@@ -407,7 +596,6 @@ async def test_resolve_job_id__from_uri_without_owner__server_error(
             owner="user",
             cluster_name="default",
             reverse="1",
-            limit="1",
         )
         raise web.HTTPError()
 
@@ -426,46 +614,34 @@ async def test_resolve_job_id__from_uri_without_owner__server_error(
 
 
 @pytest.mark.parametrize(
-    "uri", ["job://default/user", "job://default/user/", "job:", "job://"]
+    "uri",
+    [
+        "job:",
+        "job:/",
+        "job://",
+        "job:///",
+        "job://default",
+        "job://default/",
+        "job://default/user",
+        "job://default/user/",
+        "job://default/user/name/",
+        "job://default/org/user/name/",
+        "job:/user",
+        "job:/user/",
+        "job:/user/name/",
+        "job:name/",
+    ],
 )
-async def test_resolve_job_id__from_uri__missing_job_id(
+async def test_resolve_job_id__from_uri__invalid(
     aiohttp_server: _TestServerFactory, make_client: _MakeClient, uri: str
 ) -> None:
-
     app = web.Application()
     srv = await aiohttp_server(app)
 
     async with make_client(srv.make_url("/")) as client:
-        with pytest.raises(
-            ValueError,
-            match="Invalid job URI: missing job-id or job-name",
-        ):
+        with pytest.raises(ValueError, match="Invalid job URI: job:"):
             await resolve_job(uri, client=client, status={JobStatus.RUNNING})
-        with pytest.raises(
-            ValueError,
-            match="Invalid job URI: missing job-id or job-name",
-        ):
-            await resolve_job_ex(uri, client=client, status={JobStatus.RUNNING})
-
-
-@pytest.mark.parametrize("uri", ["job://default", "job://default/", "job:/", "job:///"])
-async def test_resolve_job_id__from_uri__missing_owner(
-    aiohttp_server: _TestServerFactory, make_client: _MakeClient, uri: str
-) -> None:
-
-    app = web.Application()
-    srv = await aiohttp_server(app)
-
-    async with make_client(srv.make_url("/")) as client:
-        with pytest.raises(
-            ValueError,
-            match="Invalid job URI: missing owner",
-        ):
-            await resolve_job(uri, client=client, status={JobStatus.RUNNING})
-        with pytest.raises(
-            ValueError,
-            match="Invalid job URI: missing owner",
-        ):
+        with pytest.raises(ValueError, match="Invalid job URI: job:"):
             await resolve_job_ex(uri, client=client, status={JobStatus.RUNNING})
 
 
@@ -770,7 +946,7 @@ async def test_calc_default_life_span_default_value(
         monkeypatch.chdir(tmp_path)
         local_conf = tmp_path / ".neuro.toml"
         # empty config
-        local_conf.write_text(toml.dumps({}))
+        local_conf.write_text(toml.dumps(cast(Dict[str, Any], {})))
         default = parse_timedelta("1d")
         assert (
             await calc_life_span(client, None, "1d", "job") == default.total_seconds()
